//=============================================================================
// AssignGen.g - Based on AssignParser,
//                  use a tree-walker to walk the AST and emit
//                  code for an imaginary stack machine...
//               Semantics assume a stack machine with tagged entries
//                  Tags:  value, address, fcn-address
//
// Author:  Bary W Pollack
// Date:    Apr. 17, 2004 - 14:00
//
//=============================================================================

// Import the necessary classes
{
import java.io.*;
import java.util.*;
import antlr.debug.misc.ASTFrame;
}


//-----------------------------------------------------------------------------
// Define a Parser, call it AETP
//-----------------------------------------------------------------------------

class AETP extends Parser;

options {
    defaultErrorHandler = false;    // Don't generate parser error handlers
    buildAST = true;                // Turn on AST tree construction
}

tokens {
    ARRAY; CALL; PROGRAM; 
}


// Define some methods and variables to use in the generated parser
{
    // Define a main
    public static void main(String[] args) {
    }

    static void sop(String s) {         //FOO
        if (true)                       //FOO
        System.out.println(s); 
    }
}

// ==================================================================

program
  : programBody 
      { sop("AETP program");
        #program = #([PROGRAM, "PROGRAM"], program); 
      }
    // end-of-file
  ;

programBody
  : "begin"!
////       statementList
    "end"!
  ;

statement
  : (IDENT (LPAREN|SEMI))=> procedureCallStatement
  | assignmentStatement
  ;

statementList
  : statement statementList
  | // nothing
  ;

assignmentStatement
  { String varx = "who"; }
  : varx=variableReference EQUALS^ expression SEMI!
      { sop("assignmentStatement: " + varx); sop(""); }
  ;

procedureCallStatement
  : ident:IDENT actualParameters SEMI!
      { System.out.println("ident=  " + ident.getText());    //FOO
        sop("procedureCallStatement: " + ident.getText()); sop("");
        #procedureCallStatement = #([CALL, "CALL"], procedureCallStatement);
      }
  ;

actualParameters
  : LPAREN! (expression (COMMA! expression)*)? RPAREN!
  ;

procedureCall
  : ident:IDENT actualParameters
      { System.out.println("ident=  " + ident.getText());    //FOO
        sop("procedureCall: " + ident.getText());
        #procedureCall = #([CALL, "CALL"], procedureCall);
      }
  ;

expression
  : multiplyingExpression ((PLUS^|MINUS^) multiplyingExpression)*
  ;

multiplyingExpression
  : signedExpression ((TIMES^|DIV^) signedExpression)*
  ;

signedExpression
  : (PLUS^|MINUS^)* primitiveElement
  ;

primitiveElement
  { String varx = ""; }
  : (IDENT LPAREN)=> procedureCall
  | varx=variableReference
  | constantValue
  | LPAREN! expression RPAREN!
  ;

variableReference returns [ String varx = "" ]
  : ident:IDENT
      { System.out.println("ident=  " + ident.getText());    //FOO
        varx = ident.getText();
      }
    ( LBRACKET! expression RBRACKET!
        { #variableReference = #([ARRAY, "ARRAY"], variableReference); }
    )?
  ;

constantValue
  : intlit:INTLIT
      { System.out.println("intlit= " + intlit.getText());    //FOO
      }
  ;

//----------------------------------------------------------------------------
// The AET tree-walker
//----------------------------------------------------------------------------

class AssignGen extends TreeParser;

options {
    importVocab = AETP;
}

/*
  These rule definitions all use the AST syntax:
    #(ROOT child.1 child.2 ... child.n);
*/

program
  { AST ast = null;}
  : #(PROGRAM (ast=expression)*)
      { E.emit("END"); }
  ;

expression returns [ AST r = _t ]
  { AST a, b = null;
    /* ..... debugging - to see what's in r .....
    System.out.print("expression. r = " + r); 
    if (r != null)
       System.out.println("\t   tSL: " + r.toStringList());
    else
       System.out.println();
    ..... end debugging ..... */
  }

  : #(CALL id01:IDENT (a=expression)*
      { E.emit("LOADF " + id01.getText());
        E.emit("CALL\t\t# call " + id01.getText());
        System.out.println(); 
        AssignParser.pSymbols.put(id01.getText(), 
                                  new Symbol(id01.getText(), Symbol.FUNCTION));
      }
    )

  | #(EQUALS a=expression b=expression)
      { E.emit("STORE\t\t# " + a + " <- " + b); 
        System.out.println(); 
      }

  | #(PLUS   a=expression b=expression)
      { E.emit("ADD    \t\t# " + a + " + " + b); 
      }

  | #(MINUS  a=expression (b=expression)?)
      { if (b != null)
           E.emit("SUB    \t\t# " + a + " - " + b); 
        else
           E.emit("NEG    \t\t# -" + a); r = a; 
      }

  | #(TIMES  a=expression b=expression)
      { E.emit("MUL    \t\t# " + a + " * " + b);
      }

  | #(DIV    a=expression b=expression)
      { E.emit("DIV    \t\t# " + a + " / " + b); 
      }

  | #(ARRAY  ida:IDENT a=expression)
      { E.emit("LOADA " + ida.getText() + " \t# " + ida.getText());
        E.emit("ARRAY\t\t# " + ida + "[" + a + "]"); r = ida; 
        AssignParser.pSymbols.put(ida.getText(), 
                                  new Symbol(ida.getText(), Symbol.ARRAYREF));
      }

  | id:IDENT
      { r = id; 
        E.emit("LOADA " + id.getText() + " \t# " + id.getText()); 
        AssignParser.pSymbols.put(id.getText(), 
                                  new Symbol(id.getText(), Symbol.VARIABLE));
      }

  | num:INTLIT
      { r = num;
        E.emit("LOADI " + num.getText() + " \t# " + num.getText()); 
        AssignParser.pSymbols.put(num.getText(), 
                                  new Symbol(num.getText(), Symbol.INTEGER));
      }
  ;

// EOF
